#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

#define NODATA     "\0\0\0"



// The node structure.
struct node {
  char data[3];
  struct node *left;
  struct node *right;
};



// Function prototypes.
void parse(struct node *root, bool input[], int pos, int end, int offset);
void addnode(struct node *root, char binary[], char data[2]);
void freenode(struct node *node);





int main(int argc, char *argv[]) {
  // Check for incorrect usage
  if (argc != 3) {
    printf("Usage: shadowhuffman_2 <romname> <offset>");
    exit(1);
  }

  // Copy arguments to other variables
  char *romname = argv[1];
  int offset = strtol(argv[2], NULL, 0);

  // Open the file.
  FILE *rom;
  rom = fopen(romname, "rb");
  if (!rom) {
    printf("Error: %s not found", romname);
    exit(1);
  }

  // Get to the proper location.
  fseek(rom, offset, SEEK_SET);

  // Establish the root node.
  struct node *root;
  root = (struct node *) malloc(sizeof(struct node));
  memset(&(root->data), '\0', 3);
  root->left = NULL;
  root->right = NULL;

  // Construct the tree.
  addnode(root, "0000000", "ha");
  addnode(root, "00000010", "ed");
  addnode(root, "000000110", "bl");
  addnode(root, "000000111", "|");
  addnode(root, "000001", "in");
  addnode(root, "000010000", "Do");
  addnode(root, "000010001", "We");
  addnode(root, "000010010", "iv");
  addnode(root, "000010011", " e");
  addnode(root, "00001010", "d");
  addnode(root, "00001011", "un");
  addnode(root, "0000110", "me");
  addnode(root, "0000111", "it");
  addnode(root, "00010000", "ti");
  addnode(root, "000100010", "7");
  addnode(root, "000100011", "8");
  addnode(root, "0001001", " m");
  addnode(root, "00010100000", ":");
  addnode(root, "00010100001", "_");
  addnode(root, "0001010001", "g.");
  addnode(root, "000101001", "ig");
  addnode(root, "000101010", "uy");
  addnode(root, "000101011", "rt");
  addnode(root, "0001011000", "sp");
  addnode(root, "0001011001", "So");
  addnode(root, "000101101", "N");
  addnode(root, "00010111", "O");
  addnode(root, "0001100", "or");
  addnode(root, "00011010", "ut");
  addnode(root, "000110110", "w ");
  addnode(root, "000110111", "v");
  addnode(root, "000111000", "K");
  addnode(root, "000111001", "'t");
  addnode(root, "00011101", "I'");
  addnode(root, "0001111", " i");
  addnode(root, "00100000", "'");
  addnode(root, "001000010", "Ma");
  addnode(root, "001000011", "sh");
  addnode(root, "0010001", "ve");
  addnode(root, "00100100", "nt");
  addnode(root, "00100101", " p");
  addnode(root, "00100110", "?\x0A");
  addnode(root, "00100111", "ca");
  addnode(root, "0010100", " o");
  addnode(root, "0010101000", "U");
  addnode(root, "0010101001", "io");
  addnode(root, "0010101010", "y,");
  addnode(root, "0010101011", "nc");
  addnode(root, "00101011", "ow");
  addnode(root, "00101100", "fo");
  addnode(root, "001011010", "rs");
  addnode(root, "0010110110", " C");
  addnode(root, "0010110111", " W");
  addnode(root, "00101110", "e.");
  addnode(root, "00101111", "'s");
  addnode(root, "0011000", "u");
  addnode(root, "001100100", "ul");
  addnode(root, "001100101", "am");
  addnode(root, "00110011", "no");
  addnode(root, "001101", "th");
  addnode(root, "0011100", "o ");
  addnode(root, "0011101000", "ak");
  addnode(root, "0011101001", " M");
  addnode(root, "001110101", "bu");
  addnode(root, "00111011", "a");
  addnode(root, "0011110", " h");
  addnode(root, "00111110", "as");
  addnode(root, "001111110", "No");
  addnode(root, "001111111", "M");
  addnode(root, "01000", "e ");
  addnode(root, "010010", "ou");
  addnode(root, "01001100", "ee");
  addnode(root, "01001101", "00");
  addnode(root, "01001110", "b");
  addnode(root, "01001111", "oo");
  addnode(root, "0101000", " w");
  addnode(root, "0101001000", "ts");
  addnode(root, "010100100100", "\x07");
  addnode(root, "0101001001010", "\"");
  addnode(root, "01010010010110", "?");
  addnode(root, "010100100101110", "\x7F");
  addnode(root, "0101001001011110", "\x1F");
  addnode(root, "0101001001011111", "\x1E");
  addnode(root, "01010010011", "!");
  addnode(root, "0101001010", "od");
  addnode(root, "0101001011", "ia");
  addnode(root, "01010011", "ea");
  addnode(root, "01010100", "hi");
  addnode(root, "0101010100", "gu");
  addnode(root, "0101010101", "ny");
  addnode(root, "010101011", "ay");
  addnode(root, "01010110", "do");
  addnode(root, "01010111", "nd");
  addnode(root, "01011000", "f ");
  addnode(root, "01011001", "D");
  addnode(root, "010110100", "tt");
  addnode(root, "010110101", "3");
  addnode(root, "010110110", "4");
  addnode(root, "0101101110", "mu");
  addnode(root, "0101101111", " H");
  addnode(root, "01011100", "s");
  addnode(root, "01011101", "ro");
  addnode(root, "0101111000", "mi");
  addnode(root, "0101111001", "vi");
  addnode(root, "010111101", "s.");
  addnode(root, "01011111", "S");
  addnode(root, "011000000", "ry");
  addnode(root, "011000001", "e!");
  addnode(root, "011000010", "ai");
  addnode(root, "011000011", "we");
  addnode(root, "0110001", "at");
  addnode(root, "0110010", "!\x0A");
  addnode(root, "01100110", "se");
  addnode(root, "0110011100", "pr");
  addnode(root, "0110011101", "j");
  addnode(root, "0110011110", "V");
  addnode(root, "0110011111", "rd");
  addnode(root, "0110100", "to");
  addnode(root, "01101010", "-");
  addnode(root, "01101011", "te");
  addnode(root, "011011000", "fi");
  addnode(root, "0110110010", "ns");
  addnode(root, "0110110011", "r.");
  addnode(root, "01101101", "ma");
  addnode(root, "0110111", "y ");
  addnode(root, "01110000", " I");
  addnode(root, "011100010", "ac");
  addnode(root, "0111000110", " |");
  addnode(root, "0111000111", ")");
  addnode(root, "01110010", "..");
  addnode(root, "01110011", "m ");
  addnode(root, "011101000", "bo");
  addnode(root, "011101001", "la");
  addnode(root, "01110101", "wa");
  addnode(root, "01110110", "k ");
  addnode(root, "011101110", "ly");
  addnode(root, "011101111", "ey");
  addnode(root, "0111100000", "(");
  addnode(root, "0111100001", "ug");
  addnode(root, "011110001", "He");
  addnode(root, "01111001", "al");
  addnode(root, "0111101", "n ");
  addnode(root, "0111110000", "ik");
  addnode(root, "01111100010", "ty");
  addnode(root, "01111100011", "J");
  addnode(root, "011111001", ",");
  addnode(root, "0111110100", "op");
  addnode(root, "0111110101", "d.");
  addnode(root, "011111011", "H");
  addnode(root, "0111111", " s");
  addnode(root, "10000000", "n'");
  addnode(root, "100000010", "em");
  addnode(root, "1000000110", "if");
  addnode(root, "1000000111", "y.");
  addnode(root, "100000100", "u'");
  addnode(root, "100000101", "gh");
  addnode(root, "10000011", "r");
  addnode(root, "1000010", "on");
  addnode(root, "100001100", "wh");
  addnode(root, "1000011010", " k");
  addnode(root, "1000011011", "ok");
  addnode(root, "100001110", "el");
  addnode(root, "100001111", "ki");
  addnode(root, "100010", " t");
  addnode(root, "1000110", "an");
  addnode(root, "10001110", "ng");
  addnode(root, "10001111", "ur");
  addnode(root, "100100000", "5");
  addnode(root, "1001000010", "rm");
  addnode(root, "1001000011", "kn");
  addnode(root, "10010001", "o");
  addnode(root, "1001001000", "t!");
  addnode(root, "1001001001", "e,");
  addnode(root, "1001001010", "ev");
  addnode(root, "1001001011", " D");
  addnode(root, "100100110", "ck");
  addnode(root, "1001001110", "St");
  addnode(root, "1001001111000", ".");
  addnode(root, "1001001111001", "\x03");
  addnode(root, "1001001111010", "\x06");
  addnode(root, "1001001111011", "\x05");
  addnode(root, "10010011111", "da");
  addnode(root, "100101", "t ");
  addnode(root, "10011000", "g");
  addnode(root, "100110010", "I");
  addnode(root, "1001100110", "ol");
  addnode(root, "1001100111", "na");
  addnode(root, "1001101", "d ");
  addnode(root, "10011100", "et");
  addnode(root, "10011101", "en");
  addnode(root, "1001111", "r ");
  addnode(root, "10100000", "p");
  addnode(root, "101000010", "0 ");
  addnode(root, "101000011", "\x08\x08");
  addnode(root, "1010001", "yo");
  addnode(root, "1010010000", "0");
  addnode(root, "10100100010", "\x08");
  addnode(root, "10100100011", "ap");
  addnode(root, "101001001", "P");
  addnode(root, "101001010", "so");
  addnode(root, "101001011", "nu");
  addnode(root, "101001100", "av");
  addnode(root, "1010011010", "nn");
  addnode(root, "1010011011", "ju");
  addnode(root, "10100111", "l ");
  addnode(root, "101010000", "tr");
  addnode(root, "1010100010", " T");
  addnode(root, "1010100011", "'r");
  addnode(root, "101010010", "k");
  addnode(root, "101010011", "F");
  addnode(root, "101010100", "ta");
  addnode(root, "1010101010", " Y");
  addnode(root, "1010101011", "ht");
  addnode(root, "10101011", "le");
  addnode(root, "1010110", " y");
  addnode(root, "101011100", "h ");
  addnode(root, "101011101", "il");
  addnode(root, "10101111", "n");
  addnode(root, "1011000", "u ");
  addnode(root, "101100100", "pe");
  addnode(root, "101100101", "? ");
  addnode(root, "1011001100", "wo");
  addnode(root, "1011001101", "po");
  addnode(root, "101100111", "w");
  addnode(root, "1011010", ", ");
  addnode(root, "10110110", "y");
  addnode(root, "1011011100", "z");
  addnode(root, "1011011101", "mm");
  addnode(root, "101101111", "p ");
  addnode(root, "10111000", "ke");
  addnode(root, "10111001", " g");
  addnode(root, "1011101", ". ");
  addnode(root, "10111100", "c");
  addnode(root, "1011110100", " S");
  addnode(root, "1011110101", "pl");
  addnode(root, "101111011", " r");
  addnode(root, "10111110", "l");
  addnode(root, "10111111", " f");
  addnode(root, "110000000", "ge");
  addnode(root, "1100000010", "ba");
  addnode(root, "1100000011000", "$");
  addnode(root, "1100000011001", "=");
  addnode(root, "11000000110100", "\x10");
  addnode(root, "11000000110101", "\x11");
  addnode(root, "11000000110110", "Z");
  addnode(root, "11000000110111", "X");
  addnode(root, "11000000111", "9");
  addnode(root, "11000001", "be");
  addnode(root, "11000010", " n");
  addnode(root, "11000011", "es");
  addnode(root, "1100010", ".\x0A");
  addnode(root, "110001100", "us");
  addnode(root, "110001101", "1");
  addnode(root, "110001110", "G");
  addnode(root, "110001111", "ye");
  addnode(root, "110010", "\x0A");
  addnode(root, "1100110", " ");
  addnode(root, "110011100", "ce");
  addnode(root, "1100111010", "W");
  addnode(root, "1100111011", "im");
  addnode(root, "110011110", "2");
  addnode(root, "1100111110", "ss");
  addnode(root, "1100111111", "di");
  addnode(root, "110100000", "m");
  addnode(root, "110100001", "of");
  addnode(root, "11010001", "a ");
  addnode(root, "11010010", " d");
  addnode(root, "110100110", "B");
  addnode(root, "110100111", "wi");
  addnode(root, "110101000", "ho");
  addnode(root, "110101001", "om");
  addnode(root, "11010101", "ne");
  addnode(root, "11010110", " l");
  addnode(root, "110101110", "ra");
  addnode(root, "1101011110", "si");
  addnode(root, "1101011111", "It");
  addnode(root, "110110000", "n.");
  addnode(root, "110110001", "x");
  addnode(root, "110110010", "ic");
  addnode(root, "1101100110", "mo");
  addnode(root, "11011001110", "q");
  addnode(root, "11011001111", "rr");
  addnode(root, "1101101", " a");
  addnode(root, "11011100", "Th");
  addnode(root, "11011101", "is");
  addnode(root, "110111100", "R");
  addnode(root, "1101111010", "ds");
  addnode(root, "1101111011", "ld");
  addnode(root, "11011111", "g ");
  addnode(root, "1110000", "s ");
  addnode(root, "111000100", "li");
  addnode(root, "1110001010", "Wh");
  addnode(root, "11100010110", " J");
  addnode(root, "11100010111", "tl");
  addnode(root, "11100011", " c");
  addnode(root, "11100100", "f");
  addnode(root, "11100101", "e");
  addnode(root, "1110011000", " u");
  addnode(root, "1110011001", "If");
  addnode(root, "111001101", "ad");
  addnode(root, "11100111", "st");
  addnode(root, "111010000", "ir");
  addnode(root, "1110100010", "my");
  addnode(root, "1110100011", "ru");
  addnode(root, "11101001", "! ");
  addnode(root, "1110101", "er");
  addnode(root, "111011000", "lo");
  addnode(root, "1110110010", "Ja");
  addnode(root, "1110110011", " A");
  addnode(root, "111011010", "ri");
  addnode(root, "111011011", "de");
  addnode(root, "11101110", "A");
  addnode(root, "1110111100", "6");
  addnode(root, "1110111101", "t.");
  addnode(root, "111011111", "h");
  addnode(root, "11110000", "Yo");
  addnode(root, "111100010", "L");
  addnode(root, "111100011", "E");
  addnode(root, "11110010", "ll");
  addnode(root, "11110011", " b");
  addnode(root, "111101000", "co");
  addnode(root, "111101001", "T");
  addnode(root, "11110101", "ar");
  addnode(root, "11110110", "I ");
  addnode(root, "1111011100", "ag");
  addnode(root, "1111011101", "pa");
  addnode(root, "11110111100", "uc");
  addnode(root, "11110111101", "os");
  addnode(root, "111101111100", "*");
  addnode(root, "1111011111010", "\x04");
  addnode(root, "1111011111011000", "\x0F");
  addnode(root, "1111011111011001", "\x0E");
  addnode(root, "1111011111011010", "\x01");
  addnode(root, "1111011111011011", "/");
  addnode(root, "11110111110111", "Q");
  addnode(root, "11110111111", "ie");
  addnode(root, "111110000", "ot");
  addnode(root, "111110001", "t");
  addnode(root, "11111001000", "e'");
  addnode(root, "11111001001", "'l");
  addnode(root, "1111100101", "ec");
  addnode(root, "111110011", "go");
  addnode(root, "111110100", "C");
  addnode(root, "1111101010", "su");
  addnode(root, "1111101011", "id");
  addnode(root, "11111011000", "sa");
  addnode(root, "11111011001", "Y");
  addnode(root, "1111101101", "i");
  addnode(root, "111110111", "ch");
  addnode(root, "1111110", "re");
  addnode(root, "1111111", "he");



  // And parse!
  struct node *backup = root;
  struct node *curnode = root;
  unsigned char currentByte = fgetc(rom);
  unsigned char currentBit = 0x80;

  while (true) {
    while (strcmp(NODATA, curnode->data) == 0) {
      if ((currentByte & currentBit) == 0) {
        curnode = curnode->left;
      } else {
        curnode = curnode->right;
      }
      currentBit /= 2;
      if (currentBit == 0x0) {
        currentBit = 0x80;
        currentByte = fgetc(rom);
      }
    }

    if (strchr(curnode->data, '\x0A') == NULL) {
      printf("%s", curnode->data);
      curnode = backup;
    } else {
      printf("%s", curnode->data);
      goto freepoint;
    }
  }



  freepoint:

  // Free the tree.
  freenode(root);
  return 0;
}







// binary[] holds the directions within the tree, data[2] the data to place.
// This function will add nodes if necessary.
void addnode(struct node *root, char binary[], char data[2]) {
  struct node *curnode = root;
  int i;

  for (i = 0; i < strlen(binary); i++) {
    if (binary[i] == '0') {
      if (curnode->left == NULL) {
        curnode->left = (struct node *) malloc(sizeof(struct node));
        memset(&(curnode->left->data), '\0', 3);
        curnode->left->left = NULL;
        curnode->left->right = NULL;
      }
      curnode = curnode->left;
    } else if (binary[i] == '1') {
      if (curnode->right == NULL) {
        curnode->right = (struct node *) malloc(sizeof(struct node));
        memset(&(curnode->right->data), 0, 3);
        curnode->right->left = NULL;
        curnode->right->right = NULL;
      }
      curnode = curnode->right;
    } else {
      printf("There's something other than 0 or 1 in binary[].\n");
      exit(1);
    }
  }
  strcpy(curnode->data, data);
  return;
}





// Recursively removes a node and all subnodes.
void freenode(struct node *node) {
  if (node->left != NULL) {
    freenode(node->left);
    node->left = NULL;
  }
  if (node->right != NULL) {
    freenode(node->right);
    node->right = NULL;
  }
  free(node);
  return;
}
