// Shadowmap 1.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <math.h>

#define LITERAL_STOPBIT   1
#define LITERAL_DATABIT   2
#define PASTCOPY_SOURCE   3
#define PASTCOPY_STOPBIT  4
#define PASTCOPY_DATABIT  5
#define WHATS_NEXT        6



int main(int argc, char *argv[]) {
  // Check for incorrect usage.
  if (argc != 3) {
    printf("Usage: shadowmap <romname> <offset>\n");
    exit(1);
  }

  // Copy arguments to other variables.
  char *romname = argv[1];
  int offset = strtol(argv[2], NULL, 0);

  // Open the file.
  FILE *rom;
  rom = fopen(romname, "r");
  if (!rom) {
    printf("Error: %s not found\n", romname);
    exit(1);
  }

  // Seek to the specified offset.
  fseek(rom, offset, SEEK_SET);





  // Read the header.
  int howmuch = 0;
  howmuch += (fgetc(rom));
  howmuch += (fgetc(rom) * 256);

  int metastart = offset;
  metastart += (fgetc(rom));
  metastart += (fgetc(rom) * 256);
  metastart += 2;





  // Read the byte pool.
  int poolsize = (metastart - offset) - 4;
  unsigned char pool[poolsize];
  fread(pool, sizeof(unsigned char), poolsize, rom);





  // Get the first metabyte.
  int poolindex = 0;
  int metaindex = 0;
  int outindex = 0;
  fseek(rom, metastart, SEEK_SET);
  unsigned char metabyte;
  bool metabits[8];

  metabyte = fgetc(rom);
  metabits[0] = metabyte & 0x80;
  metabits[1] = metabyte & 0x40;
  metabits[2] = metabyte & 0x20;
  metabits[3] = metabyte & 0x10;
  metabits[4] = metabyte & 0x08;
  metabits[5] = metabyte & 0x04;
  metabits[6] = metabyte & 0x02;
  metabits[7] = metabyte & 0x01;





  // Parse!

  // SEVEN IS NOT EIGHT
  unsigned char bitlen[256] = {
    0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4,
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8
  };

  unsigned char out[howmuch];
  int mode = LITERAL_STOPBIT;
  int amount = 1;
  int source = 0;
  int srclen = 0;
  int i = 0;

  // DEBUG binarydump
  // printf("%04X   ", outindex);

  while (outindex < howmuch) {
    if (metaindex == 8) {
      metabyte = fgetc(rom);
      metabits[0] = metabyte & 0x80;
      metabits[1] = metabyte & 0x40;
      metabits[2] = metabyte & 0x20;
      metabits[3] = metabyte & 0x10;
      metabits[4] = metabyte & 0x08;
      metabits[5] = metabyte & 0x04;
      metabits[6] = metabyte & 0x02;
      metabits[7] = metabyte & 0x01;
      metaindex = 0;
    }

    switch (mode) {
      case LITERAL_STOPBIT:
        if (metabits[metaindex] == 0) {
          mode = LITERAL_DATABIT;
          // DEBUG binarydump
          // printf("0");
        } else {
          for (i = 0; i < amount; i++) {
            out[outindex] = pool[poolindex];
            outindex++;
            poolindex++;
          }
          amount = 1;
          if (outindex >= 256) {
            srclen = 8 + bitlen[outindex >> 8];
          } else {
            srclen = bitlen[outindex];
          }
          // DEBUG binarydump
          // printf("1\n%04X   ", outindex);
          mode = PASTCOPY_SOURCE;
        }
        metaindex += 1;
        continue;



      case LITERAL_DATABIT:
        amount *= 2;
        if (metabits[metaindex] == 1) {
          amount += 1;
        }
        mode = LITERAL_STOPBIT;
        // DEBUG binarydump
        // printf("%d ", metabits[metaindex]);
        metaindex += 1;
        continue;



      case PASTCOPY_SOURCE:
        if (srclen > 0) {
          srclen -= 1;
          source *= 2;
          source += metabits[metaindex];
          // DEBUG binarydump
          // printf("%d", metabits[metaindex]);
          metaindex += 1;
        } else {
          mode = PASTCOPY_STOPBIT;
          // DEBUG pastcopy
          // printf("%04X: From %04X, ", outindex, source);
          // DEBUG binarydump
          // printf("\n%04X   ", outindex);
        }
        continue;



      case PASTCOPY_STOPBIT:
        if (metabits[metaindex] == 0) {
          mode = PASTCOPY_DATABIT;
          // DEBUG binarydump
          // printf("0");
        } else {
          amount += 2;
          // DEBUG pastcopy
          // printf("%02X bytes\n", amount);
          for (i = 0; i < amount; i++) {
            out[outindex] = out[source + i];
            outindex++;
          }
          amount = 1;
          source = 0;
          mode = WHATS_NEXT;
          // DEBUG binarydump
          // printf("1\n%04X   ", outindex);
        }
        metaindex += 1;
        continue;



      case PASTCOPY_DATABIT:
        amount *= 2;
        if (metabits[metaindex] == 1) {
          amount += 1;
        }
        mode = PASTCOPY_STOPBIT;
        // DEBUG binarydump
        // printf("%d ", metabits[metaindex]);
        metaindex += 1;
        continue;



      case WHATS_NEXT:
        if (metabits[metaindex] == 0) {
          mode = LITERAL_STOPBIT;
        } else {
          if (outindex >= 256) {
            srclen = 8 + bitlen[outindex >> 8];
          } else {
            srclen = bitlen[outindex];
          }
          mode = PASTCOPY_SOURCE;
        }
        // DEBUG binarydump
        // printf("%d\n%04X   ", metabits[metaindex], outindex);
        metaindex += 1;
        continue;



      default:
        printf("Error: insufficient chocolate\n");
        exit(1);
    }
  }



/*
  // Output.
  FILE *out;
  out = fopen("out.txt", "w");
  fprintf(out, "%06X   ", 0x7F0000 + ROOT);
  for (i = ROOT; i < ENDR; i++) {
    fprintf(out, "%02X ", wram[i]);
    if ((i % 8) == 5) {
      fprintf(out, "\n7F%04X   ", i+1);
    }
  }
  fprintf(out, "\n\n\n");
  fprintf(out, "BITS: %d\n", pos);
  fprintf(out, "BYTES: %f\n", (pos/8.0));

  fclose(out);
*/



/*
  // Output and end.
  for (i = 0; i < howmuch; i++) {
    printf("%02X ", out[i]);
    if ((i % 16) == 15) {
      printf("\n");
    }
  }
  printf("\n");
  fclose(rom);
*/





  // Range
  printf("------------------------------------------------------------\n");
  printf("%05X to %05X\n", offset, ftell(rom) - 1);
  printf("------------------------------------------------------------\n");

  // Background colour and the unknown byte
  printf("%02X %02X\n", out[0], out[1]);
  printf("%02X\n", out[2]);
  printf("\n");
  int j = 3;
  int numofentries = 0;

  // Floors
  printf("%02X\n", out[j]);
  numofentries = out[j];
  j += 1;
  for (i = 0; i < numofentries; i++) {
    if (out[j+1] >= 0x80) {
      printf("   %02X %02X %02X %02X %02X\n", out[j], out[j+1], out[j+2], out[j+3], out[j+4]);
      j += 5;
    } else {
      printf("   %02X %02X %02X %02X\n", out[j], out[j+1], out[j+2], out[j+3]);
      j += 4;
    }
  }
  printf("\n");

  // Walls
  printf("%02X\n", out[j]);
  numofentries = out[j];
  j += 1;
  for (i = 0; i < numofentries; i++) {
    printf("   %02X %02X %02X %02X %02X\n", out[j], out[j+1], out[j+2], out[j+3], out[j+4]);
    j += 5;
  }
  printf("\n");

  // Other
  printf("%02X\n", out[j]);
  numofentries = out[j];
  j += 1;
  for (i = 0; i < numofentries; i++) {
    printf("   %02X %02X %02X %02X %02X\n", out[j], out[j+1], out[j+2], out[j+3], out[j+4]);
    j += 5;
  }

  fclose(rom);
  return 0;
}
